from dataclasses import dataclass
from typing import Dict, Union, List
from typing_extensions import TypedDict, Literal
try:
    from typing import get_args
except ImportError:

    def get_args(annotation):
        try:
            return annotation.__args__
        except AttributeError:
            # Python 3.6
            return []

from xml.etree import ElementTree

from .data import entrez_databases

# support minimal typing up to third level of nesting
# (recursive typing not yet supported, see: https://github.com/python/typing/issues/182)
JSONType0 = Union[str, int, float, bool, None, Dict[str, any], List[any]]
JSONType1 = Union[str, int, float, bool, None, Dict[str, JSONType0], List[JSONType0]]
JSONType = Union[str, int, float, bool, None, Dict[str, JSONType1], List[JSONType1]]
DataType = Union[JSONType, ElementTree.Element]


ReturnType = Literal['json', 'xml']

_EntrezDatabaseType = Literal[
    'bioproject',
    'biosample',
    'biosystems',
    'books',
    'cdd',
    'gap',
    'dbvar',
    'epigenomics',
    'nucest',
    'gene',
    'genome',
    'gds',
    'geoprofiles',
    'nucgss',
    'homologene',
    'mesh',
    'toolkit',
    'ncbisearch',
    'nlmcatalog',
    'nuccore',
    'omia',
    'popset',
    'probe',
    'protein',
    'proteinclusters',
    'pcassay',
    'pccompound',
    'pcsubstance',
    'pubmed',
    'pmc',
    'snp',
    'sra',
    'structure',
    'taxonomy',
    'unigene',
    'unists'
]


def list_literal_values(type_annotation, modifier: lambda x: x):
    return '\n' + '\n'.join(
        [f'- {modifier(arg)}' for arg in get_args(type_annotation)]
    )


class EntrezDatabaseType:
    """The database to be used, one of:
    """


EntrezDatabase = EntrezDatabaseType()
EntrezDatabase.__supertype__ = _EntrezDatabaseType
EntrezDatabaseType.__doc__ += list_literal_values(
    _EntrezDatabaseType,
    modifier=lambda arg: f':py:obj:`\'{arg}\'` - {entrez_databases["rows"][arg]["Entrez Database"]}'
)

_CommandType = Literal[
    'neighbor',
    'neighbor_score',
    'neighbor_history',
    'acheck',
    'ncheck',
    'lcheck',
    'llinks',
    'llinkslib',
    'prlinks'
]

command_descriptions = {
    'neighbor': 'return a set of UIDs in *database* linked to the input UIDs in *database_from*.',
    'neighbor_score': 'return a set of UIDs within the same database as the input UIDs along with computed similarity scores.',
    'neighbor_history': 'post the output UIDs to the Entrez History server and returns a query_key and WebEnv corresponding to the location of the output set.',
    'acheck': 'list all links available for a set of UIDs.',
    'ncheck': 'check for the existence of links within the same database for a set of UIDs. These links are equivalent to setting *database* and *database_from* to the same value.',
    'lcheck': 'check for the existence of external links (LinkOuts) for a set of UIDs.',
    'llinks': 'for each input UID, list the URLs and attributes for the LinkOut providers that are not libraries.',
    'llinkslib': 'for each input UID, list the URLs and attributes for *all* LinkOut providers including libraries.',
    'prlinks': 'list the primary LinkOut provider for each input UID, or links directly to the LinkOut provider\'s web site for a single UID if *return_mode* is set to *ref*.'
}


class CommandType:
    """The command for the ELnk query, one of:
    """


Command = CommandType()
Command.__supertype__ = _CommandType
CommandType.__doc__ += list_literal_values(
    _CommandType,
    modifier=lambda arg: f':py:obj:`\'{arg}\'`\n\t{command_descriptions[arg]}\n'
)


_IdentifierType = Union[str, int]

Identifier = _IdentifierType


@dataclass
class Example:
    name: str
    query: 'EntrezQuery'
    uri: str


class Citation(TypedDict):
    """Dictionary with the citation input data having the following keys and values:"""
    #: Journal title, e.g. ‘science‘ or ‘proc natl acad sci u s a‘
    journal: str
    #: Year
    year: int
    #: Volume
    volume: int
    #: First page
    first_page: int
    #: Author name, e.g. ‘mann bj‘ or ‘palmenberg ac‘
    author: str
    #: Identifier that will help you recognise the citation in the results, e.g. ‘Art1‘ or ‘Art2‘
    key: str



from collections import defaultdict
from csv import DictReader
from pathlib import Path
from typing import Dict, List
from typing_extensions import TypedDict


class Table(TypedDict):
    """A table which is fast for both column and row access but quite memory-expensive"""
    columns: Dict[str, List]
    rows: Dict[str, List]


def _read_table(path: Path, index: str) -> Table:
    """Return dict where each entry is a data column"""
    table = {
        'columns': defaultdict(list),
        'rows': {}
    }
    with open(path) as f:
        reader = DictReader(f, delimiter='\t')
        for row in reader:
            for field, value in row.items():
                table['columns'][field].append(value)
            table['rows'][row[index]] = row
    return table


data_path = (Path(__file__).parent / 'data').resolve()
# https://www.ncbi.nlm.nih.gov/books/NBK25497/table/chapter2.T._entrez_unique_identifiers_ui/?report=objectonly
entrez_databases = _read_table(data_path / 'entrez_databases.tsv', index='E-utility Database Name')
entrez_database_codes = entrez_databases['columns']['E-utility Database Name']



from setuptools import setup
from setuptools import find_packages


def get_long_description(file_name):
    with open(file_name) as f:
        return f.read()


if __name__ == '__main__':
    setup(
        name='easy_entrez',
        packages=find_packages(),
        package_data={'easy_entrez': ['data/*.tsv', 'py.typed']},
        # required for mypy to work
        zip_safe=False,
        version='0.3.0',
        license='MIT',
        description='Python REST API for Entrez E-Utilities: stateless, easy to use, reliable.',
        long_description=get_long_description('README.md'),
        long_description_content_type='text/markdown',
        author='Michal Krassowski',
        author_email='krassowski.michal+pypi@gmail.com',
        url='https://github.com/krassowski/easy-entrez',
        keywords=['entrez', 'pubmed', 'e-utilities', 'ncbi', 'rest', 'api', 'dbsnp', 'literature', 'mining'],
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS',
            'Topic :: Utilities',
            'Topic :: Database',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Typing :: Typed',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9'
        ],
        install_requires=['requests', 'typing_extensions', 'dataclasses>="0.7";python_version<"3.7"'],
        extras_require={
            'with_progress_bars': ['tqdm']
        }
    )



import requests
from requests import Response
from typing import Dict, List, Optional, Union
from xml.etree import ElementTree
from copy import copy
from time import time, sleep

from .batch import supports_batches
from .types import ReturnType, DataType, EntrezDatabase, CommandType, Citation
from .queries import (
    EntrezQuery, SearchQuery, SummaryQuery, FetchQuery, LinkQuery, InfoQuery, CitationQuery, uses_query,
)


def _match_all(**kwargs):
    return ' AND '.join([
        f'{value}[{field}]'
        for field, value in kwargs.items()
    ])


class EntrezResponse:
    """The wrapper around the Entrez response."""

    def __init__(self, query: EntrezQuery, response: Response, api: 'EntrezAPI'):
        self.query: EntrezQuery = query
        self.response: Response = response
        self.api: 'EntrezAPI' = api

    @property
    def content_type(self) -> ReturnType:
        declared_type = self.response.headers['Content-Type']
        if declared_type.startswith('application/json'):
            return 'json'
        if declared_type.startswith('text/xml'):
            return 'xml'
        raise ValueError(f'Unknown content type: {declared_type}')

    @property
    def data(self) -> DataType:
        if self.content_type == 'json':
            return self.response.json()
        if self.content_type == 'xml':
            return ElementTree.fromstring(self.response.content)
        raise ValueError(f'Unknown data data {self.content_type}')

    def __repr__(self):
        query = self.query
        response = self.response
        return f'<EntrezResponse status={response.status_code} for {query.summary}>'


class EntrezAPI:
    """
    Parameters:
        tool: Name of application making the E-utility call. Value must be a string with no internal spaces.
        email: E-mail address of the E-utility user.
            Value must be a string with no internal spaces, and should be a valid e-mail address.
        api_key: Since December 2018, NCBI began enforcing the practice of using an API key
            for users that post more than 3 requests per second.
            Please see `Chapter 2 <https://www.ncbi.nlm.nih.gov/books/n/helpeutils/chapter2/>`_ for a full discussion of this policy.
        return_type: Retrieval type. Determines the format of the returned output.
        minimal_interval: The time interval (seconds) to be enforced between consecutive requests;
          by default slightly over 1/3 of a second to comply with the Entrez guidelines,
          but you may increase it if you want to be kind to others,
          or decrease it if you have an API key with an appropriate consent from Entrez.
        timeout: The timeout in seconds (default 10 seconds).
        server: The server address.
    """
    def __init__(
        self, tool: str, email: str, api_key=None,
        return_type: ReturnType = 'json',
        minimal_interval: float = 0.334,
        timeout: float = 10,
        server: str = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'
    ):
        self.server = server
        self.tool = tool
        self.email = email
        self.api_key = api_key
        self.return_type = return_type
        self.minimal_interval = minimal_interval
        self._batch_size: Optional[int] = None
        self._batch_sleep_interval: int = 3
        self._last_request_time = None
        self.timeout = timeout

    def _base_params(self) -> Dict[str, str]:
        return {
            'tool': self.tool,
            'email': self.email,
            'api_key': self.api_key,
            'retmode': self.return_type
        }

    def _request(self, query: EntrezQuery, custom_payload=None) -> EntrezResponse:
        url = f'{self.server}{query.endpoint_uri}'

        base_params = self._base_params()
        query_params = query.to_params()

        data = {
            # TODO maybe warn if overwriting?
            **base_params,
            **query_params,
            **(custom_payload or {})
        }

        current_time = time()
        if self._last_request_time is not None:
            elapsed = current_time - self._last_request_time
            if elapsed < self.minimal_interval:
                to_wait = self.minimal_interval - elapsed
                sleep(to_wait)
        self._last_request_time = current_time

        if query.method == 'get':
            response = requests.get(url, params=data, timeout=self.timeout)
        elif query.method == 'post':
            response = requests.post(url, data=data, timeout=self.timeout)
        else:
            raise ValueError(f'Incorrect query method: {query.method}')

        return EntrezResponse(query=query, response=response, api=self)

    # TODO: make entrez response a generic and provide better typing of responses
    @uses_query(SearchQuery)
    def search(
        self, term: Union[str, dict], max_results: int,
        database: EntrezDatabase = 'pubmed', min_date=None, max_date=None
    ):
        if isinstance(term, dict):
            term = _match_all(**term)

        assert not min_date and not max_date  # TODO
        query = SearchQuery(term=term, max_results=max_results, database=database)
        return self._request(query=query)

    def in_batches_of(self, size: int = 100, sleep_interval: int = 3):
        batch_mode = copy(self)
        batch_mode._batch_size = size
        batch_mode._batch_sleep_interval = sleep_interval
        return batch_mode

    @supports_batches
    @uses_query(SummaryQuery)
    def summarize(
        self, ids: List[str], max_results: int,
        database: EntrezDatabase = 'pubmed'
    ):
        self._ensure_list_like(ids)
        query = SummaryQuery(ids=ids, max_results=max_results, database=database)
        return self._request(query=query)

    @supports_batches
    @uses_query(FetchQuery)
    def fetch(
        self, ids: List[str], max_results: int,
        database: EntrezDatabase = 'pubmed', return_type: ReturnType = 'xml'
    ):
        self._ensure_list_like(ids)
        query = FetchQuery(ids=ids, max_results=max_results, database=database, return_type=return_type)
        return self._request(query=query)

    @supports_batches
    @uses_query(LinkQuery)
    def link(
        self,
        # required
        ids: List[str],
        database: EntrezDatabase,
        database_from: EntrezDatabase,
        # optional
        command: CommandType = 'neighbor'
    ):
        self._ensure_list_like(ids)
        query = LinkQuery(
            ids=ids, database=database, database_from=database_from,
            command=command
        )
        return self._request(query=query)

    @uses_query(InfoQuery)
    def get_info(self, database: EntrezDatabase = None):
        query = InfoQuery(database=database)
        return self._request(query=query)

    @uses_query(CitationQuery)
    def find_citations(self, citations: List[Citation], database='pubmed'):
        query = CitationQuery(database=database, citations=citations)
        return self._request(query=query)

    @staticmethod
    def _ensure_list_like(ids: List[str]):
        """Protect user from accidentally passing and ID, say `'142'` instead of a list,

        like `['142']` as the former would actually be interpreted as three queries,
        equivalent to: `['1', '2', '3']`.
        """
        for atomic_iterable_type in [str, bytes]:
            if isinstance(ids, atomic_iterable_type):
                raise ValueError(
                    f'Received {atomic_iterable_type.__name__} but a list-like container of identifiers was expected'
                )
